/*-
 * ~~~~~~licensing~~~~~~
 * profile-management
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * You may not use this file except in accordance with the License Terms of the Copyright
 * Holder located at: https://entaxy.ru/eula . All copyrights, all intellectual property
 * rights to the Software and any copies are the property of the Copyright Holder. Unless
 * it is explicitly allowed the Copyright Holder, the User is prohibited from using the
 * Software for commercial purposes to provide services to third parties.
 * 
 * The Copyright Holder hereby declares that the Software is provided on an "AS IS".
 * Under no circumstances does the Copyright Holder guarantee or promise that the
 * Software provided by him will be suitable or not suitable for the specific purposes
 * of the User, that the Software will meet all commercial and personal subjective
 * expectations of the User, that the Software will work properly, without technical
 * errors, quickly and uninterruptedly.
 * 
 * Under no circumstances shall the Copyright Holder or its Affiliates is not liable
 * to the User for any direct or indirect losses of the User, his expenses or actual
 * damage, including, downtime; loss of bussines; lost profit; lost earnings; loss
 * or damage to data, property, etc.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.esb.platform.base.management.core.utils;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import org.osgi.framework.BundleContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.esb.platform.base.management.core.ManagementCore;
import ru.entaxy.platform.base.support.osgi.tracker.BundleTrackerCustomizerListener;

/**
 * 
 * @author sstarovoytenkov
 *
 * @param <T> type of iterable managed objects collection, e.g. List<ManagedProfile>
 * @param <S> type of item in the collection above, e.g. ManagedProfile
 * @param <I> interface of the service to register, e.g. ProfileMBean
 * @param <C> class implementing the interface above, e.g. ProfileMBeanImpl
 */
public abstract class ManagedObjectsListener<T, S, I, C extends I> implements BundleTrackerCustomizerListener<T> {

	private static final Logger log = LoggerFactory.getLogger(ManagedObjectsListener.class);
	
	protected Map<String, ManagedObjectDescriptor<S, I, C>> managed = new HashMap<>();
	
	protected BundleContext bundleContext;
	
	protected Class<I> serviceClass;
	
	public ManagedObjectsListener(BundleContext bundleContext, Class<I> serviceClass) {
		super();
		this.bundleContext = bundleContext;
		this.serviceClass = serviceClass;
	}
	
	public Map<String, ManagedObjectDescriptor<S, I, C>> getManaged() {
		return managed;
	}

	@Override
	public void added(T managedObject) {
		Iterator<S> iter = getIterator(managedObject);
		while (iter.hasNext()) {
			S object = iter.next();
			String key = getKeyValue(object);
			if (managed.containsKey(key)) {
				managed.get(key).registration.unregister();
			}
			C service = createService(object);
			if (service == null)
				continue;
			ManagedObjectDescriptor<S, I, C> descriptor = new ManagedObjectDescriptor<>();
			descriptor.object = object;
			descriptor.service = service;
			Hashtable props = new Hashtable<>();
			props.put(ManagementCore.JMX_OBJECTNAME, getJmxObjectName(object));
			descriptor.registration = bundleContext.registerService(this.serviceClass, service, props);
			managed.put(key, descriptor);
		}
	}

	@Override
	public void removed(T managedObject) {
		Iterator<S> iter = getIterator(managedObject);
		while (iter.hasNext()) {
			S object = iter.next();
			String key = getKeyValue(object);
			if (managed.containsKey(key)) {
				managed.get(key).registration.unregister();
				managed.remove(key);
			}
		}
	}

	@Override
	public void modified(T managedObject) {
		log.debug(">> modified");
		Iterator<S> iter = getIterator(managedObject);
		while (iter.hasNext()) {
			S object = iter.next();
			String key = getKeyValue(object);
			if (managed.containsKey(key)) {
				C service = managed.get(key).service;
				if (service instanceof BundleAwareMBean) {
					((BundleAwareMBean)service).refreshBundleInfo();
					log.debug(">> refreshBundleInfo");
				} else {
					log.debug(">> NOT BundleAwareMBean");
				}
			}
		}
	}
	
	protected abstract Iterator<S> getIterator(T managedObject);
	
	protected abstract String getKeyValue(S object);
	
	protected abstract C createService(S managedObject);
	
	protected abstract String getJmxObjectName(S managedObject);
}
