/*-
 * ~~~~~~licensing~~~~~~
 * schema-impl
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * You may not use this file except in accordance with the License Terms of the Copyright
 * Holder located at: https://entaxy.ru/eula . All copyrights, all intellectual property
 * rights to the Software and any copies are the property of the Copyright Holder. Unless
 * it is explicitly allowed the Copyright Holder, the User is prohibited from using the
 * Software for commercial purposes to provide services to third parties.
 * 
 * The Copyright Holder hereby declares that the Software is provided on an "AS IS".
 * Under no circumstances does the Copyright Holder guarantee or promise that the
 * Software provided by him will be suitable or not suitable for the specific purposes
 * of the User, that the Software will meet all commercial and personal subjective
 * expectations of the User, that the Software will work properly, without technical
 * errors, quickly and uninterruptedly.
 * 
 * Under no circumstances shall the Copyright Holder or its Affiliates is not liable
 * to the User for any direct or indirect losses of the User, his expenses or actual
 * damage, including, downtime; loss of bussines; lost profit; lost earnings; loss
 * or damage to data, property, etc.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.esb.platform.runtime.core.infrastructure.schema.impl;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.Map.Entry;

import javax.wsdl.WSDLException;

import org.apache.cxf.BusException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;

import io.swagger.v3.oas.models.OpenAPI;
import io.swagger.v3.oas.models.Operation;
import io.swagger.v3.oas.models.PathItem;
import io.swagger.v3.oas.models.PathItem.HttpMethod;
import ru.entaxy.esb.platform.runtime.base.connecting.generator.Generated;
import ru.entaxy.esb.resources.EntaxyResource;
import ru.entaxy.esb.resources.EntaxyResource.RESOURCE_SCOPE;
import ru.entaxy.esb.resources.EntaxyResourceMetadata;
import ru.entaxy.esb.resources.EntaxyResourceProcessor;
import ru.entaxy.esb.resources.EntaxyResourceMetadata.MetadataSection;
import ru.entaxy.platform.base.objects.EntaxyObject;
import ru.entaxy.platform.base.support.CommonUtils;
import ru.entaxy.platform.base.support.JSONUtils;
import ru.entaxy.platform.base.support.JSONUtils.JsonBuilder.JsonArrayBuilder;
import ru.entaxy.platform.core.producer.api.EntaxyProducerUtils;
import ru.entaxy.platform.core.producer.api.ProducerResult;
import ru.entaxy.platform.core.producer.executor.generationmodel.GeneratedList;
import ru.entaxy.platform.core.support.design.openapi.EntaxyOpenApiGenerator;
import ru.entaxy.platform.core.support.design.openapi.EntaxyOpenApiGenerator.GenerationResult;
import ru.entaxy.platform.core.support.runtime.EntaxyResourceUtils;

public class OpenApiResourceProcessor implements EntaxyResourceProcessor {

	private static final Logger log = LoggerFactory.getLogger(OpenApiResourceProcessor.class);
	
	protected static final String YAML_EXTENSION = "yaml";
	protected static final String JSON_EXTENSION = "json";
	
	protected static class OpenApiPostprocessor implements EntaxyOpenApiGenerator.GenerationCallback {

		private static final String GENERATED_RESOURCES = "generatedResources";

		protected EntaxyResource entaxyResource;

		protected EntaxyResourceMetadata metadata;
		protected EntaxyResourceMetadata.MetadataSection metadataSection;
		
		protected String resourcesRootLocation = "";
		protected String resourcesLocation = "";
		
		public OpenApiPostprocessor(EntaxyResource entaxyResource) {
			this.entaxyResource = entaxyResource;
		}
		
		public void process() {

			EntaxyOpenApiGenerator generator = new EntaxyOpenApiGenerator();
			
			String url = "entaxy-resource://" + entaxyResource.getURL().replace(":", "/");
			resourcesRootLocation = entaxyResource.getURL().replace(':', '/') + "/";

			String rootFolder = UUID.randomUUID().toString();
			String resourceName = url.substring(url.lastIndexOf('/')+1);
			Path target = Paths.get(System.getProperty("karaf.home"), "temp", rootFolder, resourceName);
			File f = null;
			
			try {
				f = EntaxyResourceUtils.downloadFile(target, url);
			} catch (Exception e) {
				log.error(String.format("Error downloading resource: [%s]->[%s]", url, target.toString()), e);
				return;
			}
			if (f != null) {
				url = f.getAbsolutePath();
				generator.setRootFolder(rootFolder);
			}			
			
			generator.setUrl(url);
			generator.setCallback(this);
			GenerationResult gr = generator.generate();
			if (gr.isOk()) {
				String classesPath = gr.properties.get("classes.path").toString();
				String urlToUpload = "service-resources:" + this.resourcesLocation + "/classes";
				try {
					EntaxyResourceUtils.uploadFolder(classesPath, urlToUpload);
				} catch (IOException e) {
					log.error(String.format("FAILED upload: [%s]->[%s]", classesPath, urlToUpload), e);
				}
				
				@SuppressWarnings("unchecked")
				List<String> apiClasses = (List<String>)gr.properties.getOrDefault("classes.api", Collections.emptyList());
				@SuppressWarnings("unchecked")
				List<String> modelClasses = (List<String>)gr.properties.getOrDefault("classes.model", Collections.emptyList());
				
				log.debug("=== RESULT ===");
				
				JsonObject classes = new JsonObject();
				
				JsonArray apis = new JsonArray();
				log.debug("Api:");
				for (String s: apiClasses) {
					apis.add(s);
					log.debug("\t" + s);
				}

				JsonArray models = new JsonArray();
				log.debug("Model:");
				for (String s: modelClasses) {
					models.add(s);
					log.debug("\t" + s);
				}
				classes.addProperty("folder", "classes");
				classes.add("apis", apis);
				classes.add("models", models);
				this.metadataSection.getContent().add(OpenApiResourceMetadata.PROPERTIES.CLASSES, classes);
				
				// clean generatedResources
				metadata.getSection(EntaxyResourceMetadata.SECTION_RESOURCE.NAME).getContent().remove(GENERATED_RESOURCES);
				
				this.metadata.update();
				
				generateResources(entaxyResource, metadata);
				
				JsonArrayBuilder<?> builder = JSONUtils.JsonBuilder.create(metadata.getSection(EntaxyResourceMetadata.SECTION_RESOURCE.NAME).getContent())
						.array(GENERATED_RESOURCES, false);
				String prefix = "service-resources:" + this.resourcesLocation + "/classes/";
				for (int i=0; i<apis.size(); i++) {
					String classData = prefix + apis.get(i).getAsString().replace('.', '/') + ".class;runtime";
					builder.string(classData);
				}
				for (int i=0; i<models.size(); i++) {
					String classData = prefix + models.get(i).getAsString().replace('.', '/') + ".class;runtime";
					builder.string(classData);
				}
				
				this.metadata.update();
			}
			generator.clean();
		}
		
		@Override
		public void event(GenerationResult generationResult) {
			switch (generationResult.currentPhase) {
			case CHECK:
				if (!generationResult.isFinished() && generationResult.isCurrentPhaseOk()) {
					OpenAPI api = (OpenAPI)generationResult.properties.get("api");
					generationResult.generator.setApiTitleConverted(
							api.getInfo().getTitle().replace(' ', '_').replace('-', '_').replace('.', '_')
									.toLowerCase());
					generationResult.generator.setApiVersionConverted(
							api.getInfo().getVersion().replace('.', '_'));
					generationResult.generator.setApiPackage(
						"entaxy.runtime.service." 
						+ generationResult.generator.getApiTitleConverted() 
						+ ".v_" + generationResult.generator.getApiVersionConverted() );
					processOpenApi(api);
				}
				break;
			default:
			}			
		}
		
		protected void processOpenApi(OpenAPI api) {
			this.metadata = entaxyResource.getMetadata(true);
			this.metadataSection = this.metadata.getSection(OpenApiResourceMetadata.SECTION_NAME);
			this.metadataSection.getContent().remove("error");
			JsonObject info = new JsonObject();
			info.addProperty("title", api.getInfo().getTitle());
			info.addProperty("version", api.getInfo().getVersion());
			// JsonArray operations = new JsonArray();
			JsonObject operationData = new JsonObject();
			for (Entry<String, PathItem> entry: api.getPaths().entrySet()) {
				JsonObject pathData = new JsonObject();
				Map<HttpMethod, Operation> operationMap = entry.getValue().readOperationsMap();
				for (Entry<HttpMethod, Operation> oper: operationMap.entrySet()) {
					pathData.addProperty(oper.getKey().name(), oper.getValue().getOperationId());
				}
				operationData.add(entry.getKey(), pathData);
			}
			this.metadataSection.getContent().add(OpenApiResourceMetadata.PROPERTIES.INFO, info);
			this.metadataSection.getContent().add(OpenApiResourceMetadata.PROPERTIES.OPERATIONS, operationData);
			
			this.resourcesLocation = (this.resourcesRootLocation
					+ "/" + api.getInfo().getTitle().replace(' ', '_')
					+ "/" + api.getInfo().getVersion().replace(' ', '_'))
						.replace("//", "/");
			
			this.metadataSection.getContent().addProperty(WSDLResourceMetadata.PROPERTIES.SERVICE_RESOURCES_ROOT, this.resourcesLocation);
		}

		protected void generateResources(EntaxyResource resource, EntaxyResourceMetadata meta) {
			// '{"@LIFECYCLE":["general"], "*": {"@SKIP": ["deploy+"]}}'
			String instructions = EntaxyProducerUtils.instructions()
					.lifecycle("general")
					.any().skip("deploy+")
					.getInstructionsString();
			
			MetadataSection section = meta.getSection(OpenApiResourceMetadata.SECTION_NAME);
			if (section == null)
				return;
			
			JsonObject metaContent = section.getContent();

			JsonObject object = new JsonObject();
			object.addProperty(EntaxyObject.FIELDS.FACTORY_ID, "service-resources-openapi");
			
			JsonObject properties = new JsonObject();
			properties.add("service", metaContent.deepCopy());
			properties.addProperty("parentResource", resource.getURL());
			
			object.add(EntaxyObject.FIELDS.PROPERTIES, properties);

			
			ProducerResult result = ServiceHelper.getInstance().getEntaxyProducerService()
				.produce(object, instructions);
			
			Map<String, List<Exception>> exceptions = result.getAllExceptions();
			if (!exceptions.isEmpty()) {
				log.error("ERRORS WHILE PROCESSING OPENAPI [generateResources]:");
				for (String key: exceptions.keySet()) {
					for (Exception e: exceptions.get(key))
						log.error("ERROR in command [" + key + "]", e);
				}
			} else {
				// collect generated resources
				List<Generated> generatedResources = result.findResultObject(GeneratedList.class)
						.stream()
						.filter(g -> (
								( g != null )
								&& g.getProperties().containsKey("provider")
								&& g.getProperties().containsKey("location")
								&& g.getProperties().containsKey("scope")))
						.collect(Collectors.toList());
				List<String> resourcesData = new ArrayList<>();
				for (Generated g: generatedResources) {
					String provider = g.getProperties().getOrDefault("provider", "").toString();
					if (!CommonUtils.isValid(provider))
						continue;
					String location = g.getProperties().getOrDefault("location", "").toString();
					if (!CommonUtils.isValid(location))
						continue;
					EntaxyResource.RESOURCE_SCOPE scope = RESOURCE_SCOPE.DESIGN;
					try {
						scope = RESOURCE_SCOPE.valueOf(g.getProperties().getOrDefault("acope", "design").toString().toUpperCase());
					} catch (Exception e) {
						// use default value
					}
					resourcesData.add(provider+":"+location+";"+scope.name());
				}
				
				if (resourcesData.isEmpty())
					return;
				
				JsonArrayBuilder<?> arrayBuilder = JSONUtils.JsonBuilder.create(meta.getSection(EntaxyResourceMetadata.SECTION_RESOURCE.NAME).getContent())
					.array(GENERATED_RESOURCES, false);
				for (String s: resourcesData)
					arrayBuilder.string(s);
				
				meta.update();
			}
			
		}
		
	}
	
	@Override
	public InputStream preProcess(InputStream arg0, EntaxyResource arg1) {
		return null;
	}

	@Override
	public void postProcess(EntaxyResource resource) {
		if (!resource.exists())
			return;
		if (!resource.getName().endsWith("." + YAML_EXTENSION) && 
				!resource.getName().endsWith("." + JSON_EXTENSION))
			return;
		log.info("OpenApi PROCESSOR FOR: [{}]", resource.getURL());
		
		(new OpenApiPostprocessor(resource)).process();
	}

}
