/*-
 * ~~~~~~licensing~~~~~~
 * object-runtime-core
 * ==========
 * Copyright (C) 2020 - 2023 EmDev LLC
 * ==========
 * You may not use this file except in accordance with the License Terms of the Copyright
 * Holder located at: https://entaxy.ru/eula . All copyrights, all intellectual property
 * rights to the Software and any copies are the property of the Copyright Holder. Unless
 * it is explicitly allowed the Copyright Holder, the User is prohibited from using the
 * Software for commercial purposes to provide services to third parties.
 * 
 * The Copyright Holder hereby declares that the Software is provided on an "AS IS".
 * Under no circumstances does the Copyright Holder guarantee or promise that the
 * Software provided by him will be suitable or not suitable for the specific purposes
 * of the User, that the Software will meet all commercial and personal subjective
 * expectations of the User, that the Software will work properly, without technical
 * errors, quickly and uninterruptedly.
 * 
 * Under no circumstances shall the Copyright Holder or its Affiliates is not liable
 * to the User for any direct or indirect losses of the User, his expenses or actual
 * damage, including, downtime; loss of bussines; lost profit; lost earnings; loss
 * or damage to data, property, etc.
 * ~~~~~~/licensing~~~~~~
 */
package ru.entaxy.platform.objects.runtime.impl.manager;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ru.entaxy.platform.base.objects.EntaxyObject.OBJECT_RELATION;
import ru.entaxy.platform.objects.runtime.EntaxyRuntimeObject;
import ru.entaxy.platform.objects.runtime.EntaxyRuntimeRelation;
import ru.entaxy.platform.objects.runtime.EntaxyRuntimeObject.ENTAXY_RUNTIME_OBJECT_STATE;
import ru.entaxy.platform.objects.runtime.EntaxyRuntimeObjectManager.ENTAXY_RUNTIME_OBJECT_OPERATION;
import ru.entaxy.platform.objects.runtime.EntaxyRuntimeObjectManager.ENTAXY_RUNTIME_OBJECT_OPERATION_RESULT;
import ru.entaxy.platform.objects.runtime.EntaxyRuntimeObjectManager.OperationConfig;
import ru.entaxy.platform.objects.runtime.EntaxyRuntimeObjectManager.OperationResult;

public class OperationExecutorUninstall extends OperationExecutor<OperationExecutorUninstall> {

	private static final Logger log = LoggerFactory.getLogger(OperationExecutorUninstall.class);
	
	@Override
	protected Map<String, ENTAXY_RUNTIME_OBJECT_OPERATION> doGetAffected(OperationConfig operationConfig,
			EntaxyRuntimeObject runtimeObject) {
		final Map<String, ENTAXY_RUNTIME_OBJECT_OPERATION> result = new HashMap<>();

		List<EntaxyRuntimeRelation> relations = runtimeObject.getRelations(OBJECT_RELATION.AGGREGATION);

		List<EntaxyRuntimeObject> objects = relations.stream()
				.map(rel -> rel.getDependent())
				.filter(obj -> ENTAXY_RUNTIME_OBJECT_STATE.ACTIVE.equals(getObjectState(obj)))
				.collect(Collectors.toList());
		
		for (EntaxyRuntimeObject obj: objects)
			result.put(obj.getObjectFullId(), ENTAXY_RUNTIME_OBJECT_OPERATION.UNINSTALL);
		
		
		relations = runtimeObject.getIncomiingRelations(OBJECT_RELATION.DEPENDENCY);
		relations.addAll(runtimeObject.getIncomiingRelations(OBJECT_RELATION.USE));

		objects = relations.stream()
				.map(rel -> rel.getDependent())
				.filter(obj -> ENTAXY_RUNTIME_OBJECT_STATE.ACTIVE.equals(getObjectState(obj)))
				.collect(Collectors.toList());
		
		for (EntaxyRuntimeObject obj: objects)
			result.put(obj.getObjectFullId(), ENTAXY_RUNTIME_OBJECT_OPERATION.STOP);
		
		return result;
	}
	
	@Override
	protected OperationResultImpl doExecute(OperationConfig operationConfig, EntaxyRuntimeObject runtimeObject) {

		// uninstall all compostion-dependent objects, not colocated with the current
		List<EntaxyRuntimeObject> parts = runtimeObject.getRelations(OBJECT_RELATION.COMPOSITION).stream()
				.map(rel -> rel.getDependent())
				.filter(obj -> !runtimeObject.getContainer().getId().equals(obj.getContainer().getId()))
				.collect(Collectors.toList());
		
		for (EntaxyRuntimeObject part: parts) {
			OperationResult result = objectManager.uninstallForced(part.getObjectFullId());
			if (ENTAXY_RUNTIME_OBJECT_OPERATION_RESULT.FAILED.equals(result.getResult()))
				return OperationResultImpl.createFailed().message(
						"FAILED stopping [" + part.getObjectFullId() + "]; ORIGINAL MESSAGE: " + result.getMessage());
		}
		
		// process (uninstall or stop) all affected objects
		for (Entry<String, ENTAXY_RUNTIME_OBJECT_OPERATION> entry: operationConfig.getAffected().entrySet()) {
			
			String affectedId = entry.getKey();
			ENTAXY_RUNTIME_OBJECT_OPERATION operation = entry.getValue();
			
			OperationConfigImpl config = new OperationConfigImpl(affectedId);
			config.forced(true);
			
			OperationResult result = objectManager.execute(operation, config);
			if (ENTAXY_RUNTIME_OBJECT_OPERATION_RESULT.FAILED.equals(result.getResult()))
				return OperationResultImpl.createFailed().message(
						"FAILED operation [" + operation.name() + "] on [" + affectedId + "]; ORIGINAL MESSAGE: " + result.getMessage());
		}
		
		try {
			getBundleController(runtimeObject).uninstallBundle(runtimeObject.getContainer().getName());
			return OperationResultImpl.createDone();
		} catch (Exception e) {
			log.error("ERROR stoping [" + runtimeObject.getObjectFullId() + "]", e);
			return OperationResultImpl.createFailed().message("ERROR uninstalling [" + runtimeObject.getObjectFullId() 
				+ "]; ORIGINAL MESSAGE: " + e.getMessage());
		}
		
	}

}
